@@ -2,6 +2,7 @@ import asyncio
 import os
 import time
 import functools
+from pprint import pprint
 
 import aiohttp
 from dotenv import load_dotenv
@@ -9,10 +10,12 @@ from dotenv import load_dotenv
 from getwowdataasync.urls import urls
 from getwowdataasync.helpers import *
 
+
 class WowApi:
     """Instantiate with WowApi.create('region') to use its methods to query the API."""
+
     @classmethod
-    async def create(cls, region: str):
+    async def create(cls, region: str, locale: str = "en_US"):
         """Gets an instance of WowApi with an access token, region, and aiohttp session.
 
         Args:
@@ -23,11 +26,32 @@ class WowApi:
         """
         self = WowApi()
         self.region = region
-        timeout = aiohttp.ClientTimeout(connect=5, sock_read=60, sock_connect=5)
+        self.locale = locale
+        timeout = aiohttp.ClientTimeout(connect=5, sock_read=180, sock_connect=5)
         self.session = aiohttp.ClientSession(raise_for_status=True, timeout=timeout)
         self.access_token = await self._get_access_token()
         return self
 
+    def retry(retries: int = 5):
+        """Calls the passed in function {retries} times."""
+
+        def retries_wrapper(func):
+            @functools.wraps(func)
+            async def wrapper(*args, **kwargs):
+                start = time.monotonic()
+                for _ in range(retries):
+                    try:
+                        json = await func(*args, **kwargs)
+                        return json
+                    except aiohttp.ClientConnectionError as e:
+                        print(f"{func.__name__} {e}")
+                    except aiohttp.ClientResponseError as e:
+                        print(f"{func.__name__} {e.status}")
+
+            return wrapper
+
+        return retries_wrapper
+
     @retry()
     async def _get_access_token(
         self, wow_api_id: str = None, wow_api_secret: str = None
@@ -37,7 +61,6 @@ class WowApi:
         Args:
             wow_api_id (str): The id from a battle.net api client.
             wow_api_secret (str): The secret from a battle.net api client.
-            retries (int): The number of times the request will be retried on failure.
         """
         load_dotenv()
         id = os.environ["wow_api_id"]
@@ -54,7 +77,6 @@ class WowApi:
             response = await response.json()
             return response["access_token"]
 
-
     @retry()
     async def _fetch_get(self, url_name: str, ids: dict = {}) -> dict:
         """Preforms a aiohttp get request for the given url_name from urls.py. Accepts ids for get methods.
@@ -63,7 +85,6 @@ class WowApi:
             url_name (str): The name of a url from urls.py
             ids (dict): The ids that need to be send with the revelant url_name.
                 Such as some item_id.
-            retries (int): The number of times the request will be retried on failure.
 
         Returns:
             The content from an endpoint as binary or a dict depending if the request
@@ -73,19 +94,29 @@ class WowApi:
             **{
                 "access_token": self.access_token,
                 "namespace": f"dynamic-{self.region}",
+                "locale": self.locale,
             }
         }
         if url_name in [
-            "repice_icon", "profession_icon", "item_icon", 
-            "profession_index", "profession_skill_tier",
-            "profession_tier_detail", "profession_icon",
-            "recipe_detail", "repice_icon", "item_classes",
-            "item_subclass", "item_set_index", "item_icon"
-            ]:
+            "repice_icon",
+            "profession_icon",
+            "item_icon",
+            "profession_index",
+            "profession_skill_tier",
+            "profession_tier_detail",
+            "profession_icon",
+            "recipe_detail",
+            "repice_icon",
+            "item_classes",
+            "item_subclass",
+            "item_set_index",
+            "item_icon",
+        ]:
             params = {
                 **{
                     "access_token": self.access_token,
                     "namespace": f"static-{self.region}",
+                    "locale": self.locale,
                 }
             }
 
@@ -96,11 +127,8 @@ class WowApi:
             json["Date"] = response.headers["Date"]
             return json
 
-
     @retry()
-    async def _fetch_search(
-        self, url_name: str, extra_params: dict
-    ) -> dict:
+    async def _fetch_search(self, url_name: str, extra_params: dict) -> dict:
         """Preforms a aiohttp get request for the given url_name from urls.py. Accepts extra_params for search methods.
 
         Args:
@@ -111,16 +139,19 @@ class WowApi:
         Returns:
             The search results json parsed into a dict.
         """
-        params = {
-            "access_token": self.access_token,
-            "namespace": f"static-{self.region}",
-        }
-
-        if url_name == 'search_realm':
+        if url_name == "search_realm":
             params = {
-            "access_token": self.access_token,
-            "namespace": f"dynamic-{self.region}",
-        }
+                "access_token": self.access_token,
+                "namespace": f"dynamic-{self.region}",
+                "locale": self.locale,
+            }
+
+        elif url_name == "search_item":
+            params = {
+                "access_token": self.access_token,
+                "namespace": f"static-{self.region}",
+                "locale": self.locale,
+            }
 
         search_params = {
             **params,
@@ -130,50 +161,77 @@ class WowApi:
         async with self.session.get(
             urls[url_name].format(region=self.region), params=search_params
         ) as response:
-            json = await response.json()
+            resp_json = await response.json()
+            start = time.monotonic()
+            tasks = []
             if url_name == "search_item":
-                tasks = []
-                if json.get("results"):
-                    for item in json["results"]:
-                        task = asyncio.create_task(
-                            self._get_item(item["key"]["href"], params=params)
-                        )
-                        tasks.append(task)
-                items = await asyncio.gather(*tasks)
-                json = {}
-                json["items"] = items
-            json["Date"] = response.headers["Date"]
-            return json
+                item_json = {"items": []}
+            elif url_name == "search_realm":
+                realm_json = {"realms": []}
+
+            if resp_json.get("results"):
+                for item in resp_json["results"]:
+                    task = asyncio.create_task(self._get_item(item["key"]["href"]))
+                    tasks.append(task)
+                    if len(tasks) == 100:
+                        end = time.monotonic()
+                        elapsed = end - start
+                        print(f'100 requests executed in {elapsed}')
+                        if elapsed < 1:
+                            time.sleep(1 - elapsed)
+                        task_results = await asyncio.gather(*tasks)
+                        if url_name == "search_item":
+                            item_json["items"] += task_results
+                        elif url_name == "search_realm":
+                            realm_json["realms"] += task_results
+                        tasks = []
+                        task_results = []
+                        start = time.monotonic()
+
+            task_results = await asyncio.gather(*tasks)
+
+            if url_name == "search_item":
+                item_json["items"] += task_results
+                item_json["Date"] = response.headers["Date"]
+                return item_json
+
+            elif url_name == "search_realm":
+                realm_json["realms"] += task_results
+                realm_json["Date"] = response.headers["Date"]
+                return realm_json
 
 
     @retry()
-    async def _get_item(self, url: str, params: dict) -> dict:
-        """Preforms a get request for an item inside an item search.
+    async def _get_item(self, url: str) -> dict:
+        """Preforms a get request.
 
-        This is a general session.get() but it is currently used in item_search()
-        to get detailed item data from the href's in the search results.
+        This is a general session.get() but it is 
+        to get detailed item data from the href's 
+        in the search results.
 
         Args:
             url (str): The url to query.
-            params (dict): The parameters needed for a successful query.
-                Such as access token and namespace.
 
         Returns:
             The json response from the url as a dict.
         """
 
+        params = {
+            "access_token": self.access_token,
+            "namespace": f"static-{self.region}",
+            "locale": self.locale,
+        }
+
         async with self.session.get(url, params=params) as item_data:
             item = await item_data.json(content_type=None)
             return item
 
-
     async def connected_realm_search(self, **extra_params: dict) -> dict:
         """Preforms a search of all realms in that region.
 
         Args:
             extra_params (dict): Parameters for refining a search request.
                 See https://develop.battle.net/documentation/world-of-warcraft/guides/search
-            retries (int): The number of times the request will be retried on failure.
 
         Returns:
             The search results as json parsed into a dict.
@@ -187,7 +245,6 @@ class WowApi:
         Args:
             extra_params (dict): Parameters for refining a search request.
                 See https://develop.battle.net/documentation/world-of-warcraft/guides/search
-            retries (int): The number of times the request will be retried on failure.
 
         Returns:
             The search results as json parsed into a dict.
@@ -217,6 +274,92 @@ class WowApi:
         ids = {"connected_realm_id": connected_realm_id}
         return await self._fetch_get(url_name, ids)
 
+    async def get_all_profession_data(self) -> list:
+        """Returns a nested dictionary of all profession data.
+
+        The structure of the returned dict is like:
+            [
+                {
+                    'name' : 'Inscription',
+                    'id' : x,
+                    'skill_tiers' : [
+                        {
+                            'name' : 'Shadowlands Inscription',
+                            'id' : y,
+                            'categories' : [
+                                {
+                                'name' : 'some name',
+                                'recipes' : [
+                                        {recipe data...},
+                                        {recipe2 data...},
+                                ]
+                                }
+                            ]
+
+                        }
+                    ]
+                }
+            ]
+        """
+        profession_tree = []
+        profession_index = await self.get_profession_index()
+        for prof in profession_index["professions"]:
+            if (
+                prof["id"] < 1000
+            ):  # only add actual professions to the tree, not weird ones like protoform synthesis
+                prof_name = prof["name"]
+                prof_id = prof["id"]
+                skill_tier_tree = []
+                profession_tree.append(
+                    {"name": prof_name, "id": prof_id, "skill_tiers": skill_tier_tree}
+                )
+                skill_tier_index = await self._get_item(prof["key"]["href"])
+                if skill_tier_index.get("skill_tiers"):
+                    for skill_tier in skill_tier_index["skill_tiers"]:
+                        skill_tier_name = skill_tier["name"]
+                        skill_tier_id = skill_tier["id"]
+                        categories_tree = []
+                        skill_tier_tree.append(
+                            {
+                                "name": skill_tier_name,
+                                "id": skill_tier_id,
+                                "categories": categories_tree,
+                            }
+                        )
+                        categories_index = await self._get_item(
+                            skill_tier["key"]["href"]
+                        )
+                        if categories_index.get("categories"):
+                            for category in categories_index["categories"]:
+                                category_name = category["name"]
+                                recipe_leaves = []
+                                categories_tree.append(
+                                    {"name": category_name, "recipes": recipe_leaves}
+                                )
+                                tasks = []
+                                recipes = []
+                                start = time.monotonic()
+                                for recipe_obj in category["recipes"]:
+                                    task = asyncio.create_task(
+                                        self._get_item(recipe_obj["key"]["href"])
+                                    )
+                                    tasks.append(task)
+                                    if len(tasks) == 100:
+                                        end = time.monotonic()
+                                        elapsed = end - start
+                                        if elapsed < 1:
+                                            time.sleep(1 - elapsed)
+                                        recipes = await asyncio.gather(*tasks)
+                                        recipe_leaves += recipes
+                                        tasks = []
+                                        recipes = []
+                                        start = time.monotonic()
+                                if len(tasks) != 0:
+                                    recipes = await asyncio.gather(*tasks)
+                                    recipe_leaves += recipes
+                                    # time.sleep(1) # A second between categories is extremely inefficient but easy to avoid 429 errors
+        return profession_tree
+
     async def get_profession_index(self) -> dict:
         """Returns the all professions."""
         url_name = "profession_index"
@@ -317,7 +460,6 @@ class WowApi:
         url_name = "wow_token"
         return await self._fetch_get(url_name)
 
-
     async def get_connected_realm_index(self) -> dict:
         """Returns a dict of all realm's names with their connected realm id."""
         url_name = "connected_realm_index"
@@ -327,15 +469,13 @@ class WowApi:
         """Closes aiohttp.ClientSession."""
         await self.session.close()
 
-
+"""
 async def main():
-    #testing junk
-    from pprint import pprint
-
+    # testing junk
     for i in range(1):
         us = await WowApi.create("us")
         start = time.time()
-        json = await us.get_connected_realm_index()
+        json = await us.connected_realm_search()
         pprint(json)
         end = time.time()
         print(end - start)
@@ -345,3 +485,4 @@ async def main():
 if __name__ == "__main__":
     asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
     asyncio.run(main())
+"""
\ No newline at end of file
